# from numpy import who
from pydub import AudioSegment
from pydub.effects import normalize
from pydub.playback import play

from mw.types import Decibels, Milliseconds

from typing import List, Optional, cast 


class StackFrame: 
    segment: AudioSegment
    # cursor: Milliseconds
    in_point: Optional[Milliseconds]
    out_point: Optional[Milliseconds]
    view_start: Milliseconds
    view_end: Milliseconds

    def __init__(self, segment: AudioSegment):
        self.segment = segment
        # self.cursor = Milliseconds(0)
        self.in_point = None
        self.out_point = None
        self.view_start = Milliseconds(0)
        self.view_end = Milliseconds( len(segment) )

    def view_length(self) -> Milliseconds:
        assert self.view_end > self.view_start
        return Milliseconds( self.view_end - self.view_start )
    
    def zoom(self, factor: float):
        pass

    def crop(self, start: Milliseconds, end: Milliseconds):
        assert end > start, "crop end must be > crop start"
        self.segment = cast(AudioSegment, self.segment[start:end])
        self.in_point = None
        self.out_point = None
        # self.cursor = Milliseconds(0)
        self.view_start = Milliseconds(0)
        self.view_end = Milliseconds(len(self.segment))

    def insert_silence(self, duration: Milliseconds, at: Milliseconds):
        assert at < len(self.segment), "Insertion point past end of sound"
        a = self.segment[0:at]
        b = self.segment[at:]
        silence = AudioSegment.silent(duration=duration)
        self.segment = a + silence + b
        self.view_start = Milliseconds(0)
        self.view_end = Milliseconds(len(self.segment))

    def bloop(self, duration: Milliseconds, at: Milliseconds):
        assert at + duration < len(self.segment)
        a = self.segment[0:at]
        b = self.segment[at + duration:]
        silence = AudioSegment.silent(duration)
        self.segment = a + silence + b

        self.view_start = Milliseconds(0)
        self.view_end = Milliseconds(len(self.segment))
    
    def normalize(self, start: Milliseconds, end: Milliseconds, level: Decibels):
        assert 0 <= start < len(self.segment)
        assert 0 <= end < len(self.segment)
        assert start < end

        a = cast(AudioSegment, self.segment[0:start])
        b = cast(AudioSegment, self.segment[start:end])
        c = cast(AudioSegment, self.segment[end:])
        b = normalize(b, level)
        
        self.segment = a + b + c


    def fade_in(self, to: Milliseconds):
        assert (0 <= to < len(self.segment))
        self.segment = self.segment.fade_in(to)

    def fade_out(self, at: Milliseconds):
        assert (0 <= at < len(self.segment))
        self.segment = self.segment.fade_out(len(self.segment) - at)

    def clip_for_view(self) -> AudioSegment:
        return cast(AudioSegment, self.segment[self.view_start:self.view_end])

    def clip(self) -> AudioSegment:
        return self.segment

    def play(self):
        play(self.segment)

    def pad(self, to_length: Milliseconds):
        to_add = len(self.segment) - to_length
        if to_add > 0:
            self.segment = self.segment + AudioSegment.silent(to_add)

        self.view_start = Milliseconds(0)
        self.view_end = Milliseconds(len(self.segment))

    def export(self, filename):
        self.segment.export(filename,format='wav')


class Stack:
    entries: List[StackFrame]

    def __init__(self, segments : List[AudioSegment]):
        self.entries = []
        for segment in segments:
            self.push_sound(segment)
        
    @property
    def top(self) -> Optional[StackFrame]:
        """Get the top of the stack safetly."""
        if len(self.entries) > 0:
            return self.entries[-1]
        else:
            return None

    def push_sound(self, segment: AudioSegment):
        print(f"Pushing audio ({len(segment)} ms) onto stack...")
        self.entries.append(StackFrame(segment=segment))

    def create_new(self, length: Milliseconds):
        n = StackFrame(segment=AudioSegment.silent(length, 48000))
        self.entries.append(n)

    def split(self, at: Milliseconds):
        assert self.top is not None, "No sound on stack"
        to_split = self.top.segment
        a = cast(AudioSegment, to_split[0:at])
        b = cast(AudioSegment, to_split[at:])
        self.entries.pop()
        self.entries.append(StackFrame(a))
        self.entries.append(StackFrame(b))

    def append(self):
        assert len(self.entries) > 1

        a = self.entries.pop().segment
        b = self.entries.pop().segment
        self.entries.append(StackFrame(a + b))

    def prepend(self):
        assert len(self.entries) > 1

        a = self.entries.pop().segment
        b = self.entries.pop().segment
        self.entries.append(StackFrame(b + a))

    def loop(self, count: int = 2):
        assert len(self.entries) > 0
        a = self.entries.pop()
        segment = cast(AudioSegment, a.segment[a.in_point or 0:a.out_point or len(a.segment)])
        self.entries.append(StackFrame(segment * count))

    def bounce(self):
        assert len(self.entries) > 1
        
        a = self.entries[-1].segment
        b = self.entries[-2].segment

        if len(a) < len(b):
            a, b = b, a

        self.entries.pop()
        self.entries.pop()

        self.entries.append(StackFrame(a.overlay(b)))


    def length(self) -> Milliseconds:
        return Milliseconds(max(list(map(lambda x: len(x.clip()), self.entries)) + [0]))





from typing import Optional

import mw
from mw.types import Milliseconds

from pydub import AudioSegment
from apeek import WaveformData, unicode_waveform

class Display:
    # view_start: Milliseconds
    # view_end:  Milliseconds
    display_width: int

    def __init__(self):
        self.display_width = 80
        # self.view_start = Milliseconds(0)
        # self.view_end = Milliseconds(1)

    def max_waveform_width(self) -> int:
        return self.display_width - 5
    
    # def view_length(self) -> Milliseconds:
    #     return Milliseconds(self.view_end - self.view_start)

    def print_width_for_length(self, length: Milliseconds, view_length: Milliseconds) -> int:
        return int(self.max_waveform_width() * length / view_length )

    def create_sized_text_waveform(self, clip: AudioSegment, height: int, view_length: Optional[Milliseconds] = None) -> str:
        clip_view_length = Milliseconds(len(clip))
        if view_length is None:
            view_length = clip_view_length
        
        bins = self.print_width_for_length(clip_view_length, view_length)

        waveform = WaveformData.create_waveform_data(clip, time_bin_count=bins)
        return unicode_waveform(waveform.value_pairs, height=height)

    def print_frame(self, index, frame: 'mw.stack.StackFrame', session_length: Milliseconds):
        waveform_txt = self.create_sized_text_waveform(frame.clip(), height=2, view_length=session_length)
        print(waveform_txt.ljust(self.max_waveform_width()) + f" {index:02}")

    def print_frame_single(self, frame: 'mw.stack.StackFrame'):
        waveform_txt = self.create_sized_text_waveform(frame.clip_for_view(), height=6)
        print(waveform_txt)

    def print_stack(self, stack: 'mw.stack.Stack'):
        if len(stack.entries) > 0:
            session_length = stack.length()
            for i, frame in enumerate(reversed(stack.entries)):
                self.print_frame(i, frame, session_length)
            print(f"Session length {stack.length() / 1000.0} sec")
        else:
            print("Stack empty")

    def print_head(self, stack: 'mw.stack.Stack'):
        if stack.top:
            self.print_ruler(stack.top)
            self.print_frame_single(stack.top)
            self.print_selection(stack.top)
        else:
            print("Stack empty")
    
    def print_ruler(self, entry: 'mw.stack.StackFrame'):
        start_time = f"{entry.view_start}"
        end_time = f"{entry.view_end} ms"
        slug = list(" " * self.print_width_for_length(entry.view_length(),
                                                      entry.view_length()))

        slug[0:len(start_time)] = start_time
        slug[-len(end_time):] = end_time
        print("".join(slug))

    def print_selection(self, entry: 'mw.stack.StackFrame'):
        slug = list(" " * self.display_width)
        in_pos = None
        out_pos = None
        view_length = Milliseconds(len(entry.clip_for_view()))
        if entry.in_point is not None:
            in_pos = self.print_width_for_length(entry.in_point, view_length) 
            slug[in_pos] = "["

        if entry.out_point is not None:
            out_pos = self.print_width_for_length(entry.out_point, view_length)
            slug[out_pos] = "]"

        if in_pos is not None and out_pos is not None:
            for i in range(in_pos + 1, out_pos):
                slug[i] = "⎯"

        # slug[self.print_width_for_length(entry.cursor, view_length)] = "⬆"
        print("".join(slug))

    def show_view_info(self):
        print(f"Display width: {self.display_width} cols")
        # print(f"View start: {self.view_start} ms")
        # print(f"View end: {self.view_end} ms")
        # print(f"ms/col: {(self.view_end - self.view_start) // self.display_width}")






""" 
__main__.py

"""

import optparse

import pydub

from mw import __version__
from mw.app import App


def print_banner():
    print(f"mw {__version__}")
    print(f"(c) 2023 Jamie Hardt. All Rights Reserved.")


def main():
    parser = optparse.OptionParser()
    parser.add_option("-e", "--exec", help="Execute command", 
                      action="append", metavar="COMMAND")
    parser.add_option("-f", "--file", help="Execute comand file",
                      action="append", metavar="FILE")

    (options, files) = parser.parse_args()
    
    app = App()
    
    print_banner()

    for file in files:
        print(f"Reading audio file {file}...")
        audio = pydub.AudioSegment.from_file(file)
        app.stack.push_sound(audio)
    
    for com_file in options.file or []:
        print(f"Executing commands in {com_file}...")
        with open(com_file, "r") as f:
            for line in f.readlines():
                app.handle_command_line(line)

    for command in options.exec or []:
        app.handle_command_line(command)

    app.run()


if __name__ == "__main__":
    main()



import inspect
from copy import deepcopy
from typing import List, Callable, Optional

import mw
from mw.types import Decibels, Milliseconds

from parsimonious.exceptions import IncompleteParseError
from parsimonious.grammar import Grammar
from parsimonious import NodeVisitor


command_grammar = Grammar(
    r"""
    command = number? ("," number)? (sep? action arglist)? (sep* "#" comment)? 
    arglist = (sep argument)*
    argument = (quoted / word)
    quoted = quote literal quote
    action = ~r"[A-z]+[A-z0-9\-]*"
    number = ~r"-?[\d]+"
    word = ~r"[^\s#]+"
    quote = "\""
    comment = ~r".*"
    literal = ~r"[^\"#]*"
    sep = ~r"\s+"
    """)

class CommandParser(NodeVisitor):
    def visit_command(self, _, visited_children):
        start_part, end_part, imperative_part, _ = visited_children

        retval = {}
        
        for start in start_part:
            retval['in_addr'] = start

        for end in end_part:
            _, retval['out_addr'] = end

        for imperative in imperative_part:
            _, retval['action'], arg_list = imperative
            retval['arguments'] = arg_list
 
        return retval
    
    def visit_arglist(self, _, visited_children) -> List[str]:
        repeating_args = visited_children

        retval = []
        for arg_form in repeating_args:
            _, arg = arg_form
            retval.append(arg[0])

        return retval

    def visit_number(self, node, _) -> int:
        return int(node.text)

    def visit_action(self, node, _) -> str:
        return node.text

    def visit_word(self, node, _) -> str:
        return node.text

    def visit_quoted(self, _, visited_children) -> str:
        _, word, _ = visited_children
        return word.text

    def generic_visit(self, node, visited_children):
        return visited_children or node


def parse_numeric(base_value: int, val: str):
    """
    A helper function for converting numeric entries from the command prompt.
    """
    if val[0] in ["+","-"] and val[1:].isdigit():
        base_value += int(val)
    elif val.isdigit():
        base_value = int(val)
    return base_value


class CommandHandler:
    """
    The command handler implements commands originating from the prompt. The 
    App parses command lines into words and then hands these to the handler's 
    _handle method. The first word is the command name, and a attribute with a 
    matching name is searched in the CommandHandler instance. If a match isn't 
    found, the "help" method is run.

    If a match IS found, the corresponding attribute is called, with the App 
    instance passed as the first parameter. If any additional words were 
    present on the command line, they are passed as strings afterward.

    The help() method iterates through all the "normal" named attributes on the 
    class and prints the docstring for each as the help text.
    """
    _effective_in: Optional[Milliseconds]
    _effective_out: Optional[Milliseconds]

    def __init__(self):
        self._parser_grammar = command_grammar
        self._parser_visitor = CommandParser()
    
    def _handle_command(self, app: 'mw.app.App', command: str): 
        
        try: 
            command_dict = self._parser_visitor.visit(
                self._parser_grammar.parse(command)
            )
        except IncompleteParseError as e:
            print(f"Error: Command could not be parsed.")
            return

        self._effective_in = None
        self._effective_out = None        
        
        if app.stack.top is not None:
            self._effective_in = app.normalize_command_time(
                command_dict.get('in_addr', app.stack.top.in_point or 0))

            self._effective_out = app.normalize_command_time(
                command_dict.get('out_addr', app.stack.top.out_point or -1))
            
            if 'in_addr' in command_dict:
                app.stack.top.in_point = self._effective_in

            if 'out_addr' in command_dict:
                app.stack.top.out_point = self._effective_out

            if self._effective_in is not None \
                and self._effective_out is not None \
                and self._effective_out < self._effective_in:
                self._effective_in, self._effective_out = \
                self._effective_out, self._effective_in

        if 'action' in command_dict.keys():
            if command_dict['action'] in self._available_commands():

                # try:
                args = command_dict.get('arguments', [])
                getattr(self, command_dict['action'])(app, *args)
                # except TypeError:
                #     print(f"Error: action {command_dict['action']} called 
                #     with incorrect argument list.")
            else:
                print(f"Error: action {command_dict['action']} " 
                      f"is not recognized.")


    def _available_commands(self) -> List[str]:
        return [f for f in dir(self) if not f.startswith("_")]

    def _partial_completion_handler(self) -> Callable[[str, int],Optional[str]]:
        def _impl_autocomplete(partial: str, state: int) -> Optional[str]:
            possible = [name for name in self._available_commands() \
                if name.startswith(partial)]
            if len(possible) > 0:
                return possible[state]
            else:
                return None

        return _impl_autocomplete

    def q(self, app: 'mw.app.App'):
        "Exit the program"
        app.should_exit = True

    def help(self, _ : 'mw.app.App'):
        "Print help"
        for f in self._available_commands(): 
            m = getattr(self, f)
            argspec = inspect.signature(m)
            if len(argspec.parameters) == 1:
                print(f"{f:15}: {m.__doc__}")
            else:
                pnames = list(argspec.parameters)[1:] 
                pnames = "[" + ",".join(pnames) + "]"
                print(f"{f} {pnames}".ljust(15) + f": {m.__doc__}")
    
    # def license(self, app: 'mw.app.App'):
    #     "Print the license"
    #     print(app.license())
 
    def stack(self, app: 'mw.app.App'):
        "Print the stack"
        app.display.print_stack(app.stack)

    def show(self, app: 'mw.app.App'):
        "Show the current sound"
        app.display.print_head(app.stack)

    def ci(self, app: 'mw.app.App'):
        "Clear in point"
        if app.stack.top:
            app.stack.top.in_point = None

        app.display.print_head(app.stack)

    def co(self, app:'mw.app.App'):
        "Clear out point"
        if app.stack.top:
            app.stack.top.out_point = None

        app.display.print_head(app.stack)
    def setw(self, app: 'mw.app.App', width = "80"):
        "Set columns width"
        app.display.display_width = int(width)
        app.display.print_head(app.stack)
    
    def dup(self, app: 'mw.app.App'):
        "Push a copy of the current sound onto the stack"
        if app.stack.top:
            sound = deepcopy(app.stack.top.segment)
            app.stack.push_sound(sound)
            app.display.print_stack(app.stack)
    
    def swap(self, app:'mw.app.App'):
        "Swap the top two sounds on the stack"
        if len(app.stack.entries) > 1:
            app.stack.entries[-1], app.stack.entries[-2] = \
                app.stack.entries[-2], app.stack.entries[-1]
        
        app.display.print_stack(app.stack)

    def pop(self, app:'mw.app.App'):
        "Pop the top sound on the stack, deleting it"
        app.stack.entries.pop()
        app.display.print_stack(app.stack)
    
    def roll(self, app:'mw.app.App', count: str = "1"):
        "Roll the stack"
        if count.isdigit() or count[0] == "-" and count[1:].isdigit():
            num = int(count)
            num = num % len(app.stack.entries)
            for _ in range(num):
                app.stack.entries = app.stack.entries[-1:] + \
                    app.stack.entries[0:-1]
        else:
            print(f"Parse error: \"{count}\" is not a number")

    def crop(self, app: 'mw.app.App',):
        "Crop the sound to the in and out points"
        if app.stack.top:
            assert self._effective_in is not None
            assert self._effective_out is not None
            app.stack.top.crop(self._effective_in, self._effective_out)
        
        app.display.print_head(app.stack)

    def silence(self, app:'mw.app.App'):
        "Insert silence at in-point"
        if app.stack.top:
            assert self._effective_out is not None
            assert self._effective_in is not None
            app.stack.top.insert_silence(
                Milliseconds(self._effective_out - self._effective_in),  
                self._effective_in)
        
        app.display.print_head(app.stack)

    def new(self, app:'mw.app.App', length = "1000"):
        "Creates a new sound of [length] milliseconds"
        app.stack.create_new(length=Milliseconds(int(length)))

    def split(self, app:'mw.app.App'):
        "Split sound"
        if app.stack.top:
            assert self._effective_in is not None
            app.stack.split(self._effective_in)
        app.display.print_stack(app.stack)

    def append(self, app:'mw.app.App'):
        "Append sound"
        if len(app.stack.entries) > 1:
            app.stack.append()
        app.display.print_stack(app.stack)

    def prepend(self, app:'mw.app.App'):
        "Prepend sound"
        if len(app.stack.entries) > 1:
            app.stack.prepend()
        app.display.print_stack(app.stack)

    def loop(self, app:'mw.app.App', count = "2"):
        "Loop sound"
        if app.stack.top:
            app.stack.loop(count=int(count))
        app.display.print_head(app.stack)
    
    def normalize(self, app:'mw.app.App', level = "0.0"):
        if app.stack.top:
            assert self._effective_in is not None
            assert self._effective_out is not None

            app.stack.top.normalize(self._effective_in, 
                                    self._effective_out, 
                                    Decibels(float(level)))

        app.display.print_head(app.stack)


    def fadein(self, app:'mw.app.App'):
        "Fade in from cilp start to in point"
        if app.stack.top:
            assert self._effective_in is not None
            app.stack.top.fade_in(self._effective_in)

        app.display.print_head(app.stack)

    def fadeout(self, app:'mw.app.App'):
        "Fade out from out point to end of file"
        if app.stack.top:
            assert self._effective_out is not None 
            app.stack.top.fade_out(self._effective_out)
        
        app.display.print_head(app.stack)

    def play(self, app:'mw.app.App'):
        "Play the sound"
        if app.stack.top:
            app.stack.top.play()
    
    def length(self, app:'mw.app.App'):
        "Print the length of the top sound"
        if app.stack.top:
            print(f"{len(app.stack.top.segment)} ms")

    def bounce(self, app: 'mw.app.App'):
        "Bounce (mix) the top sound in the stack with the sound below it"
        if len(app.stack.entries) > 1:
            app.stack.bounce()

        app.display.print_stack(app.stack)

    def bloop(self, app: 'mw.app.App'):
        "Replace audio in selection with silence"
        if app.stack.top:
            assert self._effective_in is not None
            assert self._effective_out is not None

            app.stack.top.bloop(
                Milliseconds(self._effective_out - self._effective_in),                  
                self._effective_in
            )

        app.display.print_head(app.stack)

    def export(self, app: 'mw.app.App', name: str = "out.wav"):
        "Export audio as a wav file"
        if app.stack.top:
            app.stack.top.export(name)




from mw.stack import Stack
from mw.display import Display
from mw.commands import CommandHandler

from os.path import join, split

from mw.types import Milliseconds
from typing import Optional

try:
    import gnureadline as readline
except ImportError:
    import readline


class App:
    display: Display
    stack: Stack
    command_handler: CommandHandler
    should_exit: bool

    def __init__(self):
        self.stack = Stack([])
        self.display = Display()
        self.command_handler = CommandHandler()
        self.should_exit = False
        completer = self.command_handler._partial_completion_handler()
        readline.set_completer(completer)
        readline.parse_and_bind("tab: complete")

    def get_input(self):
        selection = []

        if self.stack.top:
            if self.stack.top.in_point is not None:
                selection.append(f"[{self.stack.top.in_point}")

            if self.stack.top.out_point is not None:
                selection.append(f"{self.stack.top.out_point}]")
            
            selection = "→".join(selection)
            return input(f"{selection}> ")
        else:
            return input(f"- > ")

    def handle_command_line(self, command: str):
        self.command_handler._handle_command(self, command)

    def normalize_command_time(self, addr: int) -> Optional[Milliseconds]:
        if self.stack.top is not None:
            sound_length = Milliseconds(len(self.stack.top.segment) )
            if 0 <= addr <= sound_length:
                return Milliseconds(addr)
            elif addr > sound_length:
                return sound_length
            elif addr < 0:
                from_end = abs(addr)
                from_end = min(from_end, sound_length)
                return Milliseconds(sound_length - from_end)
        else:
            return None

    def run(self):
        # print("Type \"q\" to quit.")
        while not self.should_exit:
            command = self.get_input()
            self.handle_command_line(command)

 
